from fastapi import FastAPI, HTTPException
from fastapi.responses import StreamingResponse
from pydantic import BaseModel
from typing import List, Optional, Dict, Any, Union, Literal
from openai import OpenAI
import time
import json
import asyncio

app = FastAPI()

client = OpenAI(
    base_url="https://api-inference.huggingface.co/v1/",
    api_key="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"
)

HFModelType = Literal[
    "Qwen/Qwen2.5-72B-Instruct",
    "Qwen/QwQ-32B-Preview",
    "Qwen/Qwen2.5-Coder-32B-Instruct",
    "NousResearch/Hermes-3-Llama-3.1-8B",
    "microsoft/Phi-3.5-mini-instruct"
]


# Request/Response Models
class GenerateRequest(BaseModel):
    model: str
    prompt: str
    stream: bool = True
    options: Optional[Dict[str, Any]] = None
    context: Optional[List[int]] = None
    format: Optional[str] = None
    raw: Optional[bool] = False


class ChatMessage(BaseModel):
    role: str
    content: str
    images: Optional[List[str]] = None


class ChatRequest(BaseModel):
    model: str
    messages: List[ChatMessage]
    stream: bool = True
    options: Optional[Dict[str, Any]] = None
    format: Optional[str] = None


class EmbeddingRequest(BaseModel):
    model: str
    input: Union[str, List[str]]
    options: Optional[Dict[str, Any]] = None


def huggingface_model(model: HFModelType):
    return model


def calculate_duration_ns():
    """Return current time in nanoseconds for duration calculations"""
    return time.time_ns()


async def stream_generate_response(response):
    """Stream the generate response in Ollama format"""
    start_time = calculate_duration_ns()

    for chunk in response:
        if hasattr(chunk.choices[0].delta, 'content') and chunk.choices[0].delta.content:
            yield json.dumps({
                "model": response.model,
                "created_at": str(chunk.created),
                "response": chunk.choices[0].delta.content,
                "done": False
            }) + "\n"

    # Final response with metadata
    end_time = calculate_duration_ns()
    yield json.dumps({
        "model": response.model,
        "created_at": str(response.created),
        "done": True,
        "total_duration": end_time - start_time,
        "eval_count": response.usage.total_tokens if hasattr(response, 'usage') else 0
    }) + "\n"


async def stream_chat_response(response):
    """Stream the chat response in Ollama format"""
    start_time = calculate_duration_ns()

    for chunk in response:
        if hasattr(chunk.choices[0].delta, 'content') and chunk.choices[0].delta.content:
            yield json.dumps({
                "model": response.model,
                "created_at": str(chunk.created),
                "message": {
                    "role": "assistant",
                    "content": chunk.choices[0].delta.content
                },
                "done": False
            }) + "\n"

    # Final response with metadata
    end_time = calculate_duration_ns()
    yield json.dumps({
        "model": response.model,
        "created_at": str(response.created),
        "done": True,
        "total_duration": end_time - start_time,
        "eval_count": response.usage.total_tokens if hasattr(response, 'usage') else 0
    }) + "\n"


@app.post("/api/generate")
async def generate(request: GenerateRequest):
    try:
        response = client.chat.completions.create(
            model=huggingface_model(request.model),
            messages=[{"role": "user", "content": request.prompt}],
            temperature=request.options.get('temperature', 0.7) if request.options else 0.7,
            max_tokens=request.options.get('num_predict', 2048) if request.options else 2048,
            stream=request.stream
        )

        if request.stream:
            return StreamingResponse(
                stream_generate_response(response),
                media_type="text/event-stream"
            )
        else:
            return {
                "model": response.model,
                "created_at": str(response.created),
                "response": response.choices[0].message.content,
                "done": True,
                "total_duration": calculate_duration_ns(),
                "eval_count": response.usage.total_tokens if hasattr(response, 'usage') else 0
            }
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/api/chat")
async def chat(request: ChatRequest):
    try:
        messages = [{"role": msg.role, "content": msg.content} for msg in request.messages]
        response = client.chat.completions.create(
            model=huggingface_model(request.model),
            messages=messages,
            temperature=request.options.get('temperature', 0.7) if request.options else 0.7,
            max_tokens=request.options.get('num_predict', 2048) if request.options else 2048,
            stream=request.stream
        )

        if request.stream:
            return StreamingResponse(
                stream_chat_response(response),
                media_type="text/event-stream"
            )
        else:
            return {
                "model": response.model,
                "created_at": str(response.created),
                "message": {
                    "role": "assistant",
                    "content": response.choices[0].message.content
                },
                "done": True,
                "total_duration": calculate_duration_ns(),
                "eval_count": response.usage.total_tokens if hasattr(response, 'usage') else 0
            }
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


@app.post("/api/embeddings")
async def embeddings(request: EmbeddingRequest):
    try:
        response = client.embeddings.create(
            model=huggingface_model(request.model),
            input=request.input
        )

        return {
            "embeddings": response.data[0].embedding if isinstance(request.input, str) else [d.embedding for d in
                                                                                             response.data],
            "total_duration": calculate_duration_ns()
        }
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Model Management Endpoints
@app.get("/api/tags")
async def list_models():
    """List available models"""
    models = [
        {
            "name": "Qwen/Qwen2.5-72B-Instruct",
            "modified_at": str(time.time()),
            "size": 72000000000,
            "digest": "qwen-72b",
            "details": {
                "family": "qwen",
                "parameter_size": "72B"
            }
        },
        {
            "name": "NousResearch/Hermes-3-Llama-3.1-8B",
            "modified_at": str(time.time()),
            "size": 8000000000,
            "digest": "hermes-8b",
            "details": {
                "family": "llama",
                "parameter_size": "8B"
            }
        },
        {
            "name": "Qwen/QwQ-32B-Preview",
            "modified_at": str(time.time()),
            "size": 32000000000,
            "digest": "qwen-qwq-32b-preview",
            "details": {
                "family": "qwen",
                "parameter_size": "32B"
            }
        },
        {
            "name": "Qwen/Qwen2.5-Coder-32B-Instruct",
            "modified_at": str(time.time()),
            "size": 32000000000,
            "digest": "qwen-coder-32b",
            "details": {
                "family": "qwen",
                "parameter_size": "32B"
            }
        },
        {
            "name": "microsoft/Phi-3.5-mini-instruct",
            "modified_at": str(time.time()),
            "size": 3000000000,  # Approximated size for Phi-3.5-mini
            "digest": "phi-3.5-mini",
            "details": {
                "family": "phi",
                "parameter_size": "3B"
            }
        }
    ]
    return {"models": models}

HFModelType = Literal[
    "Qwen/Qwen2.5-72B-Instruct",
    "Qwen/QwQ-32B-Preview",
    "Qwen/Qwen2.5-Coder-32B-Instruct",
    "NousResearch/Hermes-3-Llama-3.1-8B",
    "microsoft/Phi-3.5-mini-instruct"
]

def ollama_serve():
    import uvicorn
    uvicorn.run(app, host="0.0.0.0", port=11435)

if __name__ == "__main__":
    ollama_serve()



from typing import Any, Dict, List, Optional, Sequence, Union, AsyncGenerator
from openai import OpenAI, AsyncOpenAI
from pydantic import BaseModel, Field, ConfigDict
from autogen_core.components.models import (
    ModelCapabilities,
    RequestUsage,
    LLMMessage,
    CreateResult,
    ChatCompletionTokenLogprob,
    ChatCompletionClient,
)
from autogen_core.components.tools import Tool, ToolSchema
from autogen_core.base import CancellationToken


class AILiteConfig(BaseModel):
    model: str = "Qwen/Qwen2.5-72B-Instruct"
    temperature: float = 0.7
    kwargs: Dict[str, Any] = Field(default_factory=dict)
    api_key: str = "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"
    base_url: str = "https://api-inference.huggingface.co/v1/"

    model_config = ConfigDict(protected_namespaces=())


class AILiteAutoGenClient(ChatCompletionClient):
    """OpenAI chat models for HuggingFace."""

    def __init__(
        self,
        model: str = "Qwen/Qwen2.5-72B-Instruct",
        temperature: float = 0.7,
        kwargs: Dict[str, Any] = None,
        api_key: str = "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",
        base_url: str = "https://api-inference.huggingface.co/v1/",
    ):
        self.config = AILiteConfig(
            model=model,
            temperature=temperature,
            kwargs=kwargs or {},
            api_key=api_key,
            base_url=base_url,
        )
        self.client = OpenAI(
            api_key=self.config.api_key,
            base_url=self.config.base_url,
        )
        self.async_client = AsyncOpenAI(
            api_key=self.config.api_key,
            base_url=self.config.base_url,
        )
        self._total_usage = RequestUsage(prompt_tokens=0, completion_tokens=0)
        self._actual_usage = RequestUsage(prompt_tokens=0, completion_tokens=0)

    async def create(
        self,
        messages: Sequence[Dict[str, str]],
        tools: Sequence[Tool | ToolSchema] = [],
        json_output: Optional[bool] = None,
        extra_create_args: Dict[str, Any] = {},
        cancellation_token: Optional[CancellationToken] = None,
    ) -> CreateResult:
        """Create a completion."""
        response = await self.async_client.chat.completions.create(
            model=self.config.model,
            messages=messages,
            **{**self.config.kwargs, **extra_create_args}
        )

        usage = RequestUsage(
            prompt_tokens=response.usage.prompt_tokens if response.usage else 0,
            completion_tokens=response.usage.completion_tokens if response.usage else 0,
        )
        return CreateResult(
            content=response.choices[0].message.content or "",
            finish_reason=response.choices[0].finish_reason,
            usage=usage,
            cached=False,
        )

    async def create_stream(
        self,
        messages: Sequence[Dict[str, str]],
        tools: Sequence[Tool | ToolSchema] = [],
        json_output: Optional[bool] = None,
        extra_create_args: Dict[str, Any] = {},
        cancellation_token: Optional[CancellationToken] = None,
    ) -> AsyncGenerator[Union[str, CreateResult], None]:
        """Create a streaming completion."""
        stream = await self.async_client.chat.completions.create(
            model=self.config.model,
            messages=messages,
            stream=True,
            **{**self.config.kwargs, **extra_create_args}
        )

        content_parts = []
        finish_reason = None
        usage = None

        async for chunk in stream:
            delta = chunk.choices[0].delta
            if hasattr(delta, "content") and delta.content is not None:
                content_parts.append(delta.content)
                yield delta.content

            if chunk.choices[0].finish_reason:
                finish_reason = chunk.choices[0].finish_reason

            if chunk.usage:
                usage = RequestUsage(
                    prompt_tokens=chunk.usage.prompt_tokens,
                    completion_tokens=chunk.usage.completion_tokens,
                )

        if not usage:
            usage = RequestUsage(prompt_tokens=0, completion_tokens=0)

        result = CreateResult(
            content="".join(content_parts),
            finish_reason=finish_reason,
            usage=usage,
            cached=False,
        )
        yield result

    def actual_usage(self) -> RequestUsage:
        """Get the actual usage."""
        return self._actual_usage

    def total_usage(self) -> RequestUsage:
        """Get the total usage."""
        return self._total_usage

    @property
    def capabilities(self) -> ModelCapabilities:
        """Get the capabilities."""
        return ModelCapabilities(
            completion=True,
            chat=True,
            function_calling=False,
            vision=False,
            json_output=False,
        )


from typing import Any, Dict, List, Optional, Sequence, Union, cast, Iterator, AsyncIterator, Literal
from langchain_core.language_models.chat_models import BaseChatModel
from langchain_core.messages import (
    AIMessage,
    BaseMessage,
    ChatMessage,
    FunctionMessage,
    HumanMessage,
    SystemMessage, AIMessageChunk,
)
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.callbacks import CallbackManagerForLLMRun, AsyncCallbackManagerForLLMRun
from langchain_core.pydantic_v1 import Field, SecretStr, root_validator
from openai import OpenAI, AsyncOpenAI
import json

HFModelType = Literal[
    "Qwen/Qwen2.5-72B-Instruct",
    "Qwen/QwQ-32B-Preview",
    "Qwen/Qwen2.5-Coder-32B-Instruct",
    "NousResearch/Hermes-3-Llama-3.1-8B",
    "microsoft/Phi-3.5-mini-instruct"
]


def _convert_message_to_dict(message: BaseMessage) -> dict:
    """Convert LangChain message to dictionary format."""
    message_dict = {"content": message.content}

    if isinstance(message, ChatMessage):
        message_dict["role"] = message.role
    elif isinstance(message, HumanMessage):
        message_dict["role"] = "user"
    elif isinstance(message, AIMessage):
        message_dict["role"] = "assistant"
    elif isinstance(message, SystemMessage):
        message_dict["role"] = "system"
    elif isinstance(message, FunctionMessage):
        message_dict["role"] = "function"
        message_dict["name"] = message.name
    else:
        raise ValueError(f"Got unknown message type: {message}")

    return message_dict


class AILiteLangChainLLM(BaseChatModel):
    """OpenAI chat models for HuggingFace."""

    model_name: str = Field(default="Qwen/Qwen2.5-72B-Instruct", description="The HF model to use.")
    temperature: float = Field(default=0.7, description="What sampling temperature to use.")
    model_kwargs: Dict[str, Any] = Field(default_factory=dict)
    openai_api_key: SecretStr = Field(default=SecretStr("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"))
    base_url: str = Field(default="https://api-inference.huggingface.co/v1/")
    max_tokens: Optional[int] = 8000
    n: int = 1
    streaming: bool = False
    client: Any = None
    async_client: Any = None

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.client = OpenAI(
            api_key=self.openai_api_key.get_secret_value(),
            base_url=self.base_url,
        )
        self.async_client = AsyncOpenAI(
            api_key=self.openai_api_key.get_secret_value(),
            base_url=self.base_url,
        )

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "ailite"

    @property
    def _identifying_params(self) -> Dict[str, Any]:
        """Get the identifying parameters."""
        return {
            "model_name": self.model_name,
            "temperature": self.temperature,
            **self.model_kwargs
        }

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        params = {
            "model": self.model_name,
            "messages": [_convert_message_to_dict(m) for m in messages],
            "temperature": self.temperature,
            "stream": True,
            **self.model_kwargs,
        }
        if stop:
            params["stop"] = stop
        if self.max_tokens:
            params["max_tokens"] = self.max_tokens

        for chunk in self.client.chat.completions.create(**params):
            if len(chunk.choices) == 0:
                continue
            choice = chunk.choices[0]
            delta_content = choice.delta.content or ""
            generation_info = {}
            if finish_reason := choice.finish_reason:
                generation_info["finish_reason"] = finish_reason
            chunk = ChatGenerationChunk(
                message=AIMessageChunk(content=delta_content),
                generation_info=generation_info or None,
            )
            yield chunk

    async def _astream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[AsyncCallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> AsyncIterator[ChatGenerationChunk]:
        params = {
            "model": self.model_name,
            "messages": [_convert_message_to_dict(m) for m in messages],
            "temperature": self.temperature,
            "stream": True,
            **self.model_kwargs,
        }
        if stop:
            params["stop"] = stop
        if self.max_tokens:
            params["max_tokens"] = self.max_tokens

        async for chunk in await self.async_client.chat.completions.create(**params):
            if len(chunk.choices) == 0:
                continue
            choice = chunk.choices[0]
            delta_content = choice.delta.content or ""
            generation_info = {}
            if finish_reason := choice.finish_reason:
                generation_info["finish_reason"] = finish_reason
            chunk = ChatGenerationChunk(
                message=AIMessageChunk(content=delta_content),
                generation_info=generation_info or None,
            )
            yield chunk

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        params = {
            "model": self.model_name,
            "messages": [_convert_message_to_dict(m) for m in messages],
            "temperature": self.temperature,
            "stream": False,
            "n": self.n,
            **self.model_kwargs,
        }
        if stop:
            params["stop"] = stop
        if self.max_tokens:
            params["max_tokens"] = self.max_tokens

        response = self.client.chat.completions.create(**params)

        generations = []
        for choice in response.choices:
            message = choice.message
            gen = ChatGeneration(
                message=AIMessage(content=message.content or ""),
                generation_info=dict(finish_reason=choice.finish_reason)
            )
            generations.append(gen)

        token_usage = response.usage.model_dump() if response.usage else {}

        return ChatResult(
            generations=generations,
            llm_output={
                "token_usage": token_usage,
                "model_name": self.model_name,
            }
        )

    async def _agenerate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        # For simplicity, using sync version for now
        return self._generate(messages, stop, run_manager, **kwargs)


from ._main import ai

def streamai(messages_or_prompt,*args, **kwargs):
    for x in ai(messages_or_prompt=messages_or_prompt,
                stream=True,*args, **kwargs):
        print(x.choices[0].delta.content,end = "",flush=True)

def yieldai(messages_or_prompt,*args, **kwargs):
    for x in ai(messages_or_prompt=messages_or_prompt,
                stream=False,*args, **kwargs):
        yield x.choices[0].delta.content



from typing import Any, Dict, List, Optional, Sequence, Union, cast, Literal

from llama_index.core.base.llms.types import (
    ChatMessage,
    ChatResponse,
    ChatResponseGen,
    CompletionResponse,
    CompletionResponseGen,
    ChatResponseAsyncGen,
    CompletionResponseAsyncGen,
    LLMMetadata,
    MessageRole,
)
from llama_index.core.llms.callbacks import llm_chat_callback, llm_completion_callback
from llama_index.core.llms.llm import ToolSelection
from llama_index.core.llms.function_calling import FunctionCallingLLM
from llama_index.core.callbacks import CallbackManager
from openai import OpenAI as BaseOpenAI, AsyncOpenAI, NOT_GIVEN, NotGiven

HFModelType = Literal[
    "Qwen/Qwen2.5-72B-Instruct",
    "Qwen/QwQ-32B-Preview",
    "Qwen/Qwen2.5-Coder-32B-Instruct",
    "NousResearch/Hermes-3-Llama-3.1-8B",
    "microsoft/Phi-3.5-mini-instruct"
]

from pydantic import Field

class AILiteLlamaIndexHFLLM(FunctionCallingLLM):
    model: str = Field(default="Qwen/Qwen2.5-72B-Instruct")
    temperature: float = Field(default=0.0)
    max_tokens: Optional[int] = Field(default=8000)
    additional_kwargs: Dict[str, Any] = Field(default_factory=dict)
    max_retries: int = Field(default=3)
    timeout: float = Field(default=60.0)
    api_key: str = Field(default="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")
    base_url: str = Field(default="https://api-inference.huggingface.co/v1/")
    reuse_client: bool = Field(default=True)

    def __init__(
        self,
        model: HFModelType = "Qwen/Qwen2.5-72B-Instruct",
        temperature: float = 0.0,
        max_tokens: Optional[int] = None,
        additional_kwargs: Optional[Dict[str, Any]] = None,
        callback_manager: Optional[CallbackManager] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            model=model,
            temperature=temperature,
            max_tokens=max_tokens,
            additional_kwargs=additional_kwargs or {},
            callback_manager=callback_manager,
            **kwargs
        )
        self._client = BaseOpenAI(
            base_url=self.base_url,
            api_key=self.api_key,
        )

    @property
    def metadata(self) -> LLMMetadata:
        """LLM metadata."""
        return LLMMetadata(
            context_window=4096,  # Assuming default context window
            num_output=self.max_tokens or -1,
            is_chat_model=True,
            is_function_calling_model=True,
            model_name=self.model,
        )

    @llm_completion_callback()
    def complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        """Completion endpoint."""
        messages = [{"role": "user", "content": prompt}]
        chat_response = self._client.chat.completions.create(
            messages=messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            **{**self.additional_kwargs, **kwargs}
        )
        return CompletionResponse(text=chat_response.choices[0].message.content or "")

    @llm_completion_callback()
    async def acomplete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        """Async completion endpoint."""
        messages = [{"role": "user", "content": prompt}]
        chat_response = await self._aclient.chat.completions.create(
            messages=messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            **{**self.additional_kwargs, **kwargs}
        )
        return CompletionResponse(text=chat_response.choices[0].message.content or "")

    @llm_completion_callback()
    def stream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseGen:
        """Streaming completion endpoint."""
        messages = [{"role": "user", "content": prompt}]
        chat_response = self._client.chat.completions.create(
            messages=messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=True,
            **{**self.additional_kwargs, **kwargs}
        )

        def gen() -> CompletionResponseGen:
            text = ""
            for chunk in chat_response:
                if chunk.choices:
                    delta = chunk.choices[0].delta.content
                    if delta:
                        text += delta
                        yield CompletionResponse(text=text, delta=delta)

        return gen()

    @llm_completion_callback()
    async def astream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseAsyncGen:
        """Async streaming completion endpoint."""
        messages = [{"role": "user", "content": prompt}]
        chat_response = await self._aclient.chat.completions.create(
            messages=messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=True,
            **{**self.additional_kwargs, **kwargs}
        )

        async def gen() -> CompletionResponseAsyncGen:
            text = ""
            async for chunk in chat_response:
                if chunk.choices:
                    delta = chunk.choices[0].delta.content
                    if delta:
                        text += delta
                        yield CompletionResponse(text=text, delta=delta)

        return gen()

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        """Chat endpoint."""
        openai_messages = [
            {"role": msg.role.value, "content": msg.content} for msg in messages
        ]
        response = self._client.chat.completions.create(
            messages=openai_messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            **{**self.additional_kwargs, **kwargs}
        )

        msg_content = response.choices[0].message.content or ""
        return ChatResponse(message=ChatMessage(role=MessageRole.ASSISTANT, content=msg_content))

    @llm_chat_callback()
    async def achat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        """Async chat endpoint."""
        openai_messages = [
            {"role": msg.role.value, "content": msg.content} for msg in messages
        ]
        response = await self._aclient.chat.completions.create(
            messages=openai_messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            **{**self.additional_kwargs, **kwargs}
        )

        msg_content = response.choices[0].message.content or ""
        return ChatResponse(message=ChatMessage(role=MessageRole.ASSISTANT, content=msg_content))

    @llm_chat_callback()
    def stream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseGen:
        """Streaming chat endpoint."""
        openai_messages = [
            {"role": msg.role.value, "content": msg.content} for msg in messages
        ]
        chat_response = self._client.chat.completions.create(
            messages=openai_messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=True,
            **{**self.additional_kwargs, **kwargs}
        )

        def gen() -> ChatResponseGen:
            content = ""
            for chunk in chat_response:
                if chunk.choices:
                    delta = chunk.choices[0].delta.content
                    if delta:
                        content += delta
                        yield ChatResponse(
                            message=ChatMessage(
                                role=MessageRole.ASSISTANT,
                                content=content
                            ),
                            delta=delta
                        )

        return gen()

    @llm_chat_callback()
    async def astream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseAsyncGen:
        """Async streaming chat endpoint."""
        openai_messages = [
            {"role": msg.role.value, "content": msg.content} for msg in messages
        ]
        chat_response = await self._aclient.chat.completions.create(
            messages=openai_messages,
            model=self.model,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=True,
            **{**self.additional_kwargs, **kwargs}
        )

        async def gen() -> ChatResponseAsyncGen:
            content = ""
            async for chunk in chat_response:
                if chunk.choices:
                    delta = chunk.choices[0].delta.content
                    if delta:
                        content += delta
                        yield ChatResponse(
                            message=ChatMessage(
                                role=MessageRole.ASSISTANT,
                                content=content
                            ),
                            delta=delta
                        )

        return gen()

    def _prepare_chat_with_tools(
        self,
        tools: Sequence["BaseTool"],
        user_msg: Optional[Union[str, ChatMessage]] = None,
        chat_history: Optional[List[ChatMessage]] = None,
        verbose: bool = False,
        allow_parallel_tool_calls: bool = False,
        **kwargs: Any,
    ) -> Dict[str, Any]:
        """Prepare chat with tools."""
        if isinstance(user_msg, str):
            user_msg = ChatMessage(role=MessageRole.USER, content=user_msg)

        messages = chat_history or []
        if user_msg:
            messages.append(user_msg)

        return {
            "messages": messages,
            "tools": [tool.metadata.to_openai_tool() for tool in tools],
            **kwargs,
        }

    def get_tool_calls_from_response(
        self, response: ChatResponse, error_on_no_tool_call: bool = True, **kwargs: Any
    ) -> List[ToolSelection]:
        """Get tool calls from response."""
        tool_calls = response.message.additional_kwargs.get("tool_calls", [])

        if not tool_calls and error_on_no_tool_call:
            raise ValueError("No tool calls found in response")

        selections = []
        for tool_call in tool_calls:
            selections.append(
                ToolSelection(
                    tool_name=tool_call.function.name,
                    tool_id=tool_call.id,
                    tool_kwargs=tool_call.function.arguments
                )
            )
        return selections